/*******************************************************************************
 * Copyright (c) 2005 Cognos Incorporated.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Cognos Incorporated - initial API and implementation
 *******************************************************************************/

package org.eclipse.pde.equinox.servlet.bridge.launcher;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.Hashtable;
import java.util.Map;

import javax.servlet.ServletConfig;

import org.eclipse.equinox.servlet.bridge.launcher.FrameworkLauncher;

/**
 * The FrameworkLauncher provides the logic to:
 * 1) init
 * 2) deploy
 * 3) start
 * 4) stop
 * 5) undeploy
 * 6) destroy
 * an instance of the OSGi framework. 
 * These 6 methods are provided to help manage the lifecycle and are called from outside this
 * class by the BridgeServlet. To create an extended FrameworkLauncher over-ride these methods to allow
 * custom behaviour.  
 */
public class PdeFrameworkLauncher extends FrameworkLauncher {
	String[] programArgs;
	private ClassLoader frameworkContextClassLoader;
	private URL[] bootPaths;
	private ChildFirstURLClassLoader frameworkClassLoader;

	public void init(ServletConfig servletConfig) {
		config = servletConfig;
		context = servletConfig.getServletContext();
		init();
		String progArgs = System.getProperty("org.eclipse.pde.http.programArgs.file");
		if (progArgs != null) {
			ObjectInputStream oos = null;
			try {
				oos = new ObjectInputStream(new FileInputStream(progArgs));
				try {
					programArgs = (String[]) oos.readObject();
					bootPaths = (URL[]) oos.readObject();
				} catch (ClassNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} finally {
					oos.close();
				}
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}
	
	/**
	 * init is the first method called on the FrameworkLauncher and can be used for any initial setup.
	 * The default behaviour is to do nothing.
	 */
	public void init() {
		// do nothing for now
	}

	/**
	 * destory is the last method called on the FrameworkLauncher and can be used for any final cleanup.
	 * The default behaviour is to do nothing.
	 */
	public void destroy() {
		// do nothing for now
	}

	/**
	 * deploy is used to move the OSGi framework libraries into a location suitable for execution.
	 * The default behaviour is to copy the contents of the webapps WEB-INF/platform directory
	 * to the webapps temp directory.
	 */
	public synchronized void deploy() {
	}

	/** undeploy is the reverse operation of deploy and removes the OSGi framework libraries from their
	 * execution location. Typically this method will only be called if a manual undeploy is requested in the 
	 * ServletBridge.
	 * By default, this method removes the OSGi install and also removes the workspace.
	 */
	public synchronized void undeploy() {
	}

	/** start is used to "start" a previously deployed OSGi framework
	 * The default behaviour will read launcher.ini to create a set of initial properties and
	 * use the "commandline" configuration parameter to create the equivalent command line arguments
	 * available when starting Eclipse. 
	 */
	public synchronized void start() {
		String[] args = programArgs;

		ClassLoader original = Thread.currentThread().getContextClassLoader();
		try {
			System.setProperty("osgi.framework.useSystemProperties", "false"); //$NON-NLS-1$ //$NON-NLS-2$
			frameworkClassLoader = new ChildFirstURLClassLoader(bootPaths, getClass().getClassLoader());
			Class clazz = frameworkClassLoader.loadClass(STARTER);
			Hashtable initalPropertyMap = new Hashtable();
			initalPropertyMap.put("osgi.parentClassloader", "fwk");

			Method setInitialProperties = clazz.getMethod("setInitialProperties", new Class[] {Map.class}); //$NON-NLS-1$
			setInitialProperties.invoke(null, new Object[] {initalPropertyMap});
			
			Method runMethod = clazz.getMethod("run", new Class[] {String[].class, Runnable.class}); //$NON-NLS-1$
			runMethod.invoke(null, new Object[] {args, null});
			frameworkContextClassLoader = Thread.currentThread().getContextClassLoader();
		} catch (Exception e) {
			e.printStackTrace();
			return;
		} finally {
			Thread.currentThread().setContextClassLoader(original);
		}
	}


	/**
	 * stop is used to "shutdown" the framework and make it avialable for garbage collection.
	 * The default implementation also has special handling for Apache Commons Logging to "release" any
	 * resources associated with the frameworkContextClassLoader.
	 */
	public synchronized void stop() {
		if (frameworkClassLoader == null) {
			context.log("Framework is already shutdown"); //$NON-NLS-1$
			return;
		}
		ClassLoader original = Thread.currentThread().getContextClassLoader();
		try {
			Class clazz = frameworkClassLoader.loadClass(STARTER);
			Method method = clazz.getDeclaredMethod("shutdown", (Class[]) null); //$NON-NLS-1$
			Thread.currentThread().setContextClassLoader(frameworkContextClassLoader);
			method.invoke(clazz, (Object[]) null);

			// ACL keys its loggers off of the ContextClassLoader which prevents GC without calling release. 
			// This section explicitly calls release if ACL is used.
			try {
				clazz = this.getClass().getClassLoader().loadClass("org.apache.commons.logging.LogFactory"); //$NON-NLS-1$
				method = clazz.getDeclaredMethod("release", new Class[] {ClassLoader.class}); //$NON-NLS-1$
				method.invoke(clazz, new Object[] {frameworkContextClassLoader});
			} catch (ClassNotFoundException e) {
				// ignore, ACL is not being used
			}
		} catch (Exception e) {
			context.log("Error while stopping Framework", e); //$NON-NLS-1$
			return;
		} finally {
			frameworkClassLoader = null;
			frameworkContextClassLoader = null;
			Thread.currentThread().setContextClassLoader(original);
		}
	}

}
